/**
* Zeigt durch roboterCodingTask definierte Level zum Lösen an
*/
function TaskUI(
    nodeContainer       // DOM element to occupy
) {
    nodeContainer.innerHTML = '';
    this._msgPlaceGeneral = new ScoTools.MessagePlace(true, true);
    this._msgPlaceGeneral.appendAsChild(nodeContainer);
    this._msgPlace = new ScoTools.MessagePlace(false, true);
    this._msgPlace.appendAsChild(nodeContainer);

    this._showTask('Wird geladen…');
    this._contentContainer = ScoTools.createHtml('div', nodeContainer, { 'class': 'codeAndTable' });
    this._codeView = new ClickableSourceCode();
    this._codeView.onchange = (line => { this.check(line); return false; });
    this._table = new TableOfVars();
    this._table.onaddfailed = (msg => this.addMsg(msg, ScoTools.Message.TYPE_INFO));
    this._table.onrequestformatedline = (linenumber => this._codeView.getFormatedLine(linenumber));
    this._table.onrequestreturn = (linenumber => this._codeView.line = linenumber);
    this._cout = new TextOut();
    this._cout.display = false;
    [this._codeView, this._table, this._cout].forEach(({ node }) => ScoTools.createHtml('div', this._contentContainer).appendChild(node) );

    const resetBtn = ScoTools.createHtml('button', ScoTools.createHtml('p', this._contentContainer.lastChild, { 'class': 'resetBtnPlace' }));
    ScoTools.createHtml('img', resetBtn, { alt: "⎌", src: "share/Edit-undo.svg" });
    ScoTools.createHtml(false, resetBtn, null, ' Auf Ausgangslage zurücksetzen');
    resetBtn.addEventListener('click', () => this.resetTable());

    this.startSeed = 0;
}

// Verwendete ScoTools.MessagePlace-Instanz für generelle Meldungen
TaskUI.prototype._msgPlaceGeneral = null;

// Verwendete ScoTools.MessagePlace-Instanz für aktuelle Meldung
TaskUI.prototype._msgPlace = null;

// Verwendete ClickableSourceCode-Instanz
TaskUI.prototype._codeView = null;

// Verwendete TableOfVars-Instanz
TaskUI.prototype._table = null;

// Container (DOM-Knoten) für Hauptinhalt
TaskUI.prototype._contentContainer = null;

// Sollen Instancen ohne Referenz darauf automatisch entfernt werden?
TaskUI.prototype.autoremoveGarbage = false;

// Angezeigter Java-Quellcode
Object.defineProperty(TaskUI.prototype, 'code', {
    enumerable: true,
    set(code) { this._codeView.code = code; },
    get() { return this._codeView.code; }
});

// Funktion, die beim Freischalten eines neuen Levels aufgerufen werden soll
TaskUI.prototype.onlevelunlock = function(unlockedLevelCount) { }

// Was ist gerade der aktive Schritt?
TaskUI.prototype._active = 0;
Object.defineProperty(TaskUI.prototype, 'active', {
    enumerable: true,
    get() { return this._active; },
    set(v) {
        const oldActive = this._active;
        if (!Number.isFinite(v) || ( v >= this.levels.length - 1)) {
            this.done = true;
            this._active = Infinity;
        } else
            this._active = Math.floor(Math.max(Math.min(this.levels.length - 1, v), 0));
        if (!Number.isFinite(this._active))
            this._showFinalMessage();
        else if ((this._active != oldActive) && (this._active != oldActive + 1))
            this._gotoLevel();
        else
            this._startLevel();
        if (this._active == oldActive)
            return;
        if ((typeof this.onlevelunlock == 'function') || (this.onlevelunlock instanceof Function))
            this.onlevelunlock.call(this, this._active);
    }
});

// Wurden Aufgabe bereits abgeschlossen?
TaskUI.prototype._done = false;
Object.defineProperty(TaskUI.prototype, 'done', {
    enumerable: true,
    get() { return this._done; },
    set(v) {
        if ((v ? true : false) <= this._done)
            return;
        this._done = true;
        this.addGeneralMsg('Aufgabe erfolgreich abgeschlossen', ScoTools.Message.TYPE_SUCCESS);
    }
});

// Funktion, die beim Erhalt von Strafpunkten aufgerufen werden soll
TaskUI.prototype.onnewfine = function(totalFine, additionalFine) { }

// Gesammelte Anzahl an Fehlerpunkten
TaskUI.prototype._fineCount = 0;
Object.defineProperty(TaskUI.prototype, 'fineCount', {
    enumerable: true,
    get() { return this._fineCount; },
    set(v) {
        if ((v == this._fineCount) || !Number.isFinite(v))
            return;

        const oldFine = this._fineCount;
        this._fineCount = v;
        if ((typeof this.onnewfine == 'function') || (this.onnewfine instanceof Function))
            this.onnewfine.call(this, v, v - oldFine);
    }
});

// Generierte Leveldaten
TaskUI.prototype._levels = [];
Object.defineProperty(TaskUI.prototype, 'levels', {
    enumerable: true,
    get() { return this._levels; },
    set(levels) {
        if (!(levels instanceof Array))
            return;
        this._levels = levels;
        this.active = 0;
    }
});

// Startwert für Pseudozufallszahlengenerierung
TaskUI.prototype._startSeed = 0;
Object.defineProperty(TaskUI.prototype, 'startSeed', {
    enumerable: true,
    get() { return this._startSeed; },
    set(v) {
        if ((v == this._startSeed) || !Number.isFinite(v) || (this._startSeed === v))
            return;
        this._startSeed = v;
        this._importTask(true);
    }
});

/**
* Einen Zufallszahlengenerationsfunktion für Ganzzahlen erstellen
* @return Function      Zufallszahlengenerationsfunktion
*/
TaskUI.prototype._getRandomIntFkt = function(
    useSeed     // Pseudozufallszahlengenerierung (mit Startwert this.startSeed) statt Math.random verwenden?
) {
    if (!useSeed)
        return ((min, max) => (Math.floor(Math.random() * (max - min + 1)) + min));
    let seed = this.startSeed;
    return ((min, max) => { seed = (seed * 9301 + 49297) % 233280; return Math.floor(seed / 233280 * (max - min + 1)) + min; });
}

/**
* Daten von window.TASK übernehmen
* @return undefined
*/
TaskUI.prototype._importTask = function(
    useSeed     // Pseudozufallszahlengenerierung (mit Startwert this.startSeed) statt Math.random verwenden?
) {
    if (!(window.TASK instanceof Object))
        return;
    let data = window.TASK;
    if (data instanceof Function)
        data = window.TASK.call(window.TASK, this._getRandomIntFkt(useSeed));
    if (!(data instanceof Object))
        return;
    ['notSelectable', 'methodCalls', 'classDeclarations'].forEach(key => data[key] && (this._codeView[key] = data[key]));
    if (data.code)
        this.code = data.code;
    this._table.classes = data.classes;
    if (data.steps)
        this.levels = data.steps.slice();
    this._stdSetting = data;
    this._gotoLevel();
}

/**
* Früher gespeicherte Werte wieder übernehmen
* @return undefined
*/
TaskUI.prototype.loadValues = function(
    isDone,                 // Wurde diese Aufgabe bereits abgeschlossen?
    activeStep,             // Anzahl der bereits abgeschlossen Schritte
    fineCount,              // Anzahl der gesammelten Fehlerpunkte
    seed                    // Startwert für die Pseudozufallszahlengenerierung
) {
    this.startSeed = seed;
    this._fineCount = (Number.isFinite(fineCount) ? fineCount : 0);
    this.active = isDone ? Infinity : activeStep;
}

/**
* Aufgabenstellung anzeigen
* @return undefined
*/
TaskUI.prototype._showTask = function(
    task        // Anzuzeigender HTML-Code
) {
    if (!task && this._nodeTask) {
        this._nodeTask.parentNode.removeChild(this._nodeTask);
        this._nodeTask = null;
        return;
    }
    if (this._nodeTask) {
        this._nodeTask.innerHTML = task;
        return
    }
    this._nodeTask = ScoTools.createHtml('div', null, { 'class': 'auftrag' });
    this._nodeTask.innerHTML = task;
    this._msgPlace.node.parentNode.insertBefore(this._nodeTask, this._msgPlaceGeneral.node.nextSibling);
}

/**
* Informationen über aktuelles Level ermitteln
* @return Object    Eintrag aus this.levels
*/
TaskUI.prototype.getLevelInfo = function() {
    const level = this.levels[this.active];
    return level ? level : null;
}

/**
* Die Speichertabelle und alles andere zurücksetzen
* @return undefined
*/
TaskUI.prototype.resetTable = function() {
    this._startLevel();
}

/**
* Den aktuellen Wert für das Attribut `key` ermitteln und dabei ggf. frühere Level berücksichtigen
* @return String    Standardausgabetext
*/
TaskUI.prototype._getAbsoluteLevelValue = function(
    key,            // Attributname in Leveldaten
    defaultValue,   // Optional: Rückgabewert, wenn kein Wert vergeben wurde
    index           // Optional: Gewünschter Index in this.levels (Standardwert this.active)
) {
    const levels = this.levels;
    for (index = Math.min(Number.isFinite(index) ? index : this.active, levels.length - 1); index >= 0; --index) {
        if (levels[index][key] !== undefined)
            return levels[index][key];
    }
    if (defaultValue !== undefined)
        return defaultValue;
    if (['cout'].indexOf(key) != -1)
        return '';
    if (['vars'].indexOf(key) != -1)
        return [];
    if (key == 'task')
        return 'Führe den Code aus, indem du immer die Auswirkungen der hervorgehobenen Zeile in der Speichertabelle einträgst und die nächste auszuführende Codezeile auswählst.';
    return false;
}

/**
* Ein Level starten, wenn evtl. nicht direkt vom vorherigen Level dorthin gewechselt wird
* @return undefined
*/
TaskUI.prototype._gotoLevel = function() {
    ['allowArrays', 'enableOverload', 'autoremoveGarbage', 'enableTextOut'].forEach(key => this._table[key] = this._getAbsoluteLevelValue(key, this._stdSetting[key]));
    this.autoremoveGarbage = this._getAbsoluteLevelValue('autoremoveGarbage', this._stdSetting.autoremoveGarbage);
    this._cout.display = this._getAbsoluteLevelValue('enableTextOut', this._stdSetting.enableTextOut);
    this._showTask(this._getAbsoluteLevelValue(('task', this._stdSetting.task ? this._stdSetting.task : 'Führe den Code aus, indem du immer die Auswirkungen der hervorgehobenen Zeile in der Speichertabelle einträgst und die nächste auszuführende Codezeile auswählst.')));
    this._startLevel();
}

/**
* Level neu starten oder direkt folgendes Level starten
* @return undefined
*/
TaskUI.prototype._startLevel = function() {
    this.clearMsg();
    const level = this.getLevelInfo();

    if (this._btnRestart && this._btnRestart.parentNode) {
        this._btnRestart.parentNode.replaceChild(this._contentContainer, this._btnRestart);
        this._btnRestart = null;
    }

    this._procLevelStart(level);
    if (level.task)
        this._showTask(level.task);
    if ((typeof level.start == 'function') || (level.start instanceof Function))
        level.start();
    this._codeView.line = this._getAbsoluteLevelValue('line');
    ['allowArrays', 'enableOverload'].map(key => {
        if (level[key] !== undefined)
            this._table[key] = level[key];
    });
    this._table.load(this._getAbsoluteLevelValue('vars'));
    if (level.autoremoveGarbage !== undefined)
        this.autoremoveGarbage = level.autoremoveGarbage;
    if (level.enableTextOut !== undefined)
        this._cout.display = level.enableTextOut;
    this._cout.value = this._getAbsoluteLevelValue('cout');
    const linenumber = this._codeView.line;
    const lineType = this._codeView.getLineType(linenumber);
    switch (lineType) {
        case 1:
            const className = this._codeView.getClassName(linenumber);
            this._table.addScope(this.active ? this.levels[this.active - 1].line : -1, className);
            break;
        case 2:
            this._table.addScope(this.active ? this.levels[this.active - 1].line : -1);
            break;
    }
}

/**
* Abschlussbericht anzeigen
* @return undefined
*/
TaskUI.prototype._showFinalMessage = function() {
    this.clearMsg();
    if (this._contentContainer.parentNode) {
        const btnRestart = ScoTools.createHtml('button', null, { 'class': 'fullwidth' }, 'Aufgabe erneut üben');
        btnRestart.addEventListener('click', () => {
            this._fineCount = 0;
            this._importTask();
        });
        this._contentContainer.parentNode.replaceChild(btnRestart, this._contentContainer);
        this._btnRestart = btnRestart;
    }
    this._showTask(this.createFinalReport());
}

/**
* Abschlussbericht zusammenstellen
* @return String        HTML-Code des Abschlussberichts
*/
TaskUI.prototype.createFinalReport = function() {
    if (!this.fineCount)
        return 'Hervorragend! Du hast die Aufgabe fehlerfrei gelöst.';
    if (this.fineCount == 1)
        return 'Gut gemacht: Du hast die Aufgabe gelöst und nur <strong>einen Fehlerpunkt</strong> kassiert.';
    const arrZahlen = ['', '', 'zwei', 'drei', 'vier', 'fünf', 'sechs', 'sieben', 'acht', 'neun', 'zehn', 'elf', 'zwölf'];
    if (this.fineCount < arrZahlen.length)
        return 'Du hast die Aufgabe gemeistert, dabei allerdings <strong>' + arrZahlen[this.fineCount] + ' Fehlerpunkte</strong> gesammelt.';
    return 'Du hast die Aufgabe gemeistert, dabei allerdings <strong>' + this.fineCount + ' Fehlerpunkte</strong> gesammelt.';
}

/**
* Dem Level ggf. mitteilen, dass es soeben (neu) gestartet wurde
* @return undefined
*/
TaskUI.prototype._procLevelStart = function(
    level       // Gestartetes Level
) {
    if ((level instanceof Object) && (level.onstart instanceof Function))
        level.onstart.call(level);
    if ((level instanceof Object) && (level.onstart instanceof Array))
        level.onstart.reduce((prev, checker) => (prev || ((checker instanceof Function) ? checker.call(level) : checker)), undefined);
}

/**
* Eine (weitere) Meldung dauerhaft anzeigen
* @return undefined
*/
TaskUI.prototype.addGeneralMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._msgPlaceGeneral.add(text, type);
}

/**
* Aktive Meldung löschen
* @return undefined
*/
TaskUI.prototype.clearMsg = function() {
    this._msgPlace.clear();
}

/**
* Eine Meldung vorübergehend anzeigen
* @return undefined
*/
TaskUI.prototype.addMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._msgPlace.add(text, type);
}

/**
* Fehler bei der Textausgabe an den Nutzer kommunizieren
* @return String        Fehlerbeschreibung
*/
TaskUI.prototype._tellCoutError = function(
    { start, end, actual, target, isEnd }      // Rückgabewert von TextOut.prototype.selectDiff
) {
    if (!start && !actual.length)
        return 'Textausgabe fehlt';
    if (!actual.length && !target.trim().length && (target.length > 1))
        return 'Textausgabe ist unvollständig: ' + (isEnd ? 'Am Ende' : 'Innerhalb des Textes') + ' fehlen unsichtbare Zeichen (bspw. Leerzeichen oder Zeilenumbruch)';
    if (!actual.length && !target.trim().length)
        return 'Textausgabe ist unvollständig: ' + (isEnd ? 'Am Ende' : 'Innerhalb des Textes') + ' fehlt ein unsichtbares Zeichen (bspw. Leerzeichen oder Zeilenumbruch)';
    if (!actual.length)
        return 'Textausgabe ist unvollständig';
    if (!actual.trim().length && (actual.length > 1))
        return 'Unsichtbare Zeichen (bspw. Leerzeichen oder Zeilenumbruch) ' + (isEnd ? 'am Ende' : 'innerhalb') + ' der Textausgabe müssen ' + (target.length ? 'ersetzt' : 'entfernt') + ' werden';
    if (!actual.trim().length)
        return 'Ein unsichtbares Zeichen (bspw. Leerzeichen oder Zeilenumbruch) ' + (isEnd ? 'am Ende' : 'innerhalb') + ' der Textausgabe muss ' + (target.length ? 'ersetzt' : 'entfernt') + ' werden';
    return 'Der markierte Teil der Textausgabe ' + (target.length ? 'ist falsch' : 'muss weg');
}

/**
* Bearbeitung des aktuellen Schritts korrigieren
* @return Boolean       Schritt richtig bearbeitet?
*/
TaskUI.prototype.check = function(
    nextLine        // Ausgewählte Zeilennummer
) {
    const level = this.levels[this.active + 1];
    const oldLine = this._codeView.line;
    if (!(level instanceof Object)) {
        ++this.active;
        return true;
    }
    if (Number.isFinite(level.line) && (level.line != nextLine) && ((level.altLines instanceof Array) ? (level.altLines.indexOf(nextLine) == -1) : true)) {
        this.addMsg('Das ist nicht die Zeile, in der es weitergeht…');
        ++this.fineCount;
        return false;
    }
    const type = this._codeView.getLineType(level.line);
    const diffs = level.vars ? this._table.compare(level.vars, this.autoremoveGarbage) : [];
    if (diffs && diffs.length) {
        //const lastDiff = diffs.pop();
        //this.addMsg('Fehler in Speichertabelle: ' + (diffs.length ? (diffs.join(', ') + ' und ') : '') + lastDiff);
        this.addMsg('Fehler in Speichertabelle: ' + diffs[0] + ((diffs.length > 1) ? ', …' : ''));
        ++this.fineCount;
        return false;
    }
    const textDiff = this._cout.display ? this._cout.selectDiff(this._getAbsoluteLevelValue('cout', undefined, this.active + 1)) : null;
    if (textDiff) {
        this.addMsg(this._tellCoutError(textDiff));
        this._cout.focus();
        ++this.fineCount;
        return false;
    }
    ++this.active;
    return true;
}
